#include <cell/pad.h>
#include <cell/rtc.h>

#include "inc/exports.h"
#include "misc.h"
#include "mem.h"
#include "blitting.h"
#include "vsh_menu.h"

#ifdef GAME_MENU
#include "pad_hook.h"
#endif

#ifdef DEBUG
#include "network.h"
#endif

#ifdef HAVE_SSCROLLER
#include "scroller.h"
#endif

#ifdef HAVE_STARFIELD
#include "starfield.h"
#endif

#define SYSCALL8_OPCODE_PS3MAPI 0x7777
#define PS3MAPI_OPCODE_GET_ALL_PROC_PID 0x0021
#define PS3MAPI_OPCODE_GET_PROC_NAME_BY_PID 0x0022
#define PS3MAPI_OPCODE_LOAD_PROC_MODULE 0x0044

// vsh menu specific methods
void check_for_booted_game(void);
void check_for_exited_game(void);
int32_t load_plugin_to_game(uint8_t *prx_path, void *arg, uint32_t arg_size);

// vsh menu data
static uint8_t plugin_running, auto_game_load;
enum current_proc {IN_VSH = 1, IN_GAME = 2};
static char current_auto_setting[128];

////////////////////////////////////////////////////////////////////////
// BLITTING
bool menu_running = 0;          // menu off(0) or on(1)
static int8_t line = 0;         // current line into menu, init 0 (entry 1:)
static int8_t view = 0;         // menu view, init 0 (main view)
#define MAX_MENU_LINE 3
#define MAX_MENU_VIEW 3
enum view_title {MAIN_VIEW, SCREENSHOT_VIEW, GAME_PLUGIN_OPTIONS_VIEW};
static int8_t view_menu[MAX_MENU_VIEW] = {3, 3, 3};
const char *entry_str[MAX_MENU_VIEW][MAX_MENU_LINE] = {
{
	"Open screenshot menu",
	"Open game menu plugin options",
	"Unload plugin",
},
{ // screenshot menu
  "Back to main view",
  "screenshot(XMB)",
  "screenshot(XMB+Menu)",
},
{ // game plugin options menu
  "Back to main view",
  "Disable automatic game plugin loading",
  "Enable automatic game plugin loading",
},
};

/***********************************************************************
* draw a frame (in xmb)
***********************************************************************/
static void draw_frame(void)
{
	int32_t i;
  char tempstr[128];

	// all 32bit colors are ARGB, the framebuffer format
	set_background_color(0xFF000000);     // black, opaque
	set_foreground_color(0xFFFFFFFF);     // white, opaque

	// fill background with background color
	draw_background();

  #ifdef HAVE_STARFIELD
  draw_stars();       // to keep them under text lines
  #endif

  #ifdef HAVE_PNG_BG
	// draw background from png
  draw_png(0, 0, 0, 0, 0, 720, 400);
  #endif

	// draw menu entry list
	for(i = 0; i < view_menu[view]; i++)
  {
    i == line ? set_foreground_color(0xFFFF0000) : set_foreground_color(0xFFFFFFFF);
    print_text(10, 10 + (FONT_H * (i + 1)), entry_str[view][i]);
  }

  #ifdef HAVE_SSCROLLER
  // testing sine in a scroller
  draw_text(330);
  move_text();
  #endif

  // menu specific printing
  set_foreground_color(0xFFFFFFFF);
  if (view == MAIN_VIEW) {
	  print_text(300, 8, "Main Menu");
  } else if (view == SCREENSHOT_VIEW) {
	  print_text(300, 8, "Screenshot Menu");
    set_foreground_color(0xFF00FF00);
    sprintf(tempstr, "Note: You may also take a screenshot by holding down [Start] and pressing\n[Select+R1]");
    print_text(10, 10 + (FONT_H * (i + 2)), tempstr);
  } else if (view == GAME_PLUGIN_OPTIONS_VIEW) {
	  print_text(300, 8, "Game Menu Plugin Options Menu");
    sprintf(tempstr, "Auto load game plugin: %s", current_auto_setting);
    print_text(10, 10 + (FONT_H * (i + 2)), tempstr);
    if (auto_game_load == 0) {
      set_foreground_color(0xFF00FF00);
      sprintf(tempstr, "Hold down [Start] and press [Select+R3]\nto manually load the game menu plugin", current_auto_setting);
      print_text(10, 10 + (FONT_H * (i + 4)), tempstr);
    }
  }
}

/***********************************************************************
* stop vsh menu
***********************************************************************/
static void stop_VSH_Menu(void)
{
  menu_running = 0;     // menu off
  #ifdef HAVE_SYS_FONT
  font_finalize();      // unbind renderer and kill font-instance
  #endif
  destroy_heap();       // free heap memory
  rsx_fifo_pause(0, 0); // continue rsx rendering
  start_stop_vsh_pad(1);
}

/***********************************************************************
* start game menu
***********************************************************************/
static int32_t start_VSH_Menu(void)
{
  int32_t ret, mem_size;

  line = 0;                                                         // menu on first entry in list
  rsx_fifo_pause(0, 1);
  #ifdef HAVE_SYS_FONT
  mem_size = ((((CANVAS_W * CANVAS_H * sizeof(uint32_t)) * 2) + (FONT_CACHE_MAX * 32 * 32)) + MB(2)) / MB(1);
  #else
  mem_size =(((CANVAS_W * CANVAS_H * sizeof(uint32_t)) * 2)/*fg & bg*/ + MB(2))/*leave some headroom*/ / MB(1)/*convert to multiple of page size 1MB*/;
  #endif
  ret = create_heap(mem_size);
  if (ret != 0) {
    return(-1);
  }
  ret = init_graphic();                                             // initialize game menu graphic
  if (ret != 0) {
    return(-1);
  }
  #ifdef HAVE_SYS_FONT
  set_font(20, 20, 1, 1);                                           // set font(char w/h = 20 pxl, line-weight = 1 pxl, distance between chars = 1 pxl)
  #endif
  #ifdef HAVE_PNG_BG
  ret = load_png_bitmap(0, PNG_BG_PATH);     // load a png background if you like
  if (ret != 0) {
    return(-1);
  }
  #endif
  start_stop_vsh_pad(0);
  menu_running = 1;                                                 // set menu_running on
  return(0);
}

/***********************************************************************
* execute action depending on line selected
***********************************************************************/
static void do_menu_action(void)
{
  switch(view)
  {
    case 0:                     // "Main Menu"
      switch(line)
      {
        case 0:                 // "Menu Entry 1: Open screenshot menu"
          view = 1;
          line = 0;
          break;
        case 1:                 // "Menu Entry 2: Open game plugin options menu"
          view = 2;
          line = 0;
          break;
        case 2:                 // "Menu Entry 3: Unload plugin"
          stop_VSH_Menu();
          plugin_running = 0;
          break;
      }
      break;
    case 1:                     // "Screenshot Menu"
      switch(line)
      {
        case 0:                 // "Menu Entry 1: Back to main menu"
          view = line = 0;
          break;
        case 1:                 // "Menu Entry 2: screenshot(XMB)"
          screenshot(0);
          stop_VSH_Menu();
          break;
        case 2:                 // "Menu Entry 3: screenshot(XMB+Menu)"
          screenshot(1);
          stop_VSH_Menu();
          break;
      }
      break;
    case 2:                     // "Game Plugin Options Menu"
      switch(line)
      {
        case 0:                 // "Menu Entry 1: Back to main menu"
          view = line = 0;
          break;
        case 1:                 // "Menu Entry 2: Disable automatic game plugin loading"
          auto_game_load = 0;
          sprintf(current_auto_setting, "DISABLED");
          break;
        case 2:                 // "Menu Entry 3: Enable automatic game plugin loading"
          auto_game_load = 1;
          sprintf(current_auto_setting, "ENABLED");
          break;
      }
      break;
  }
}

/***********************************************************************
* vsh menu main
***********************************************************************/
int32_t vsh_menu(void)
{
  #ifdef DEBUG
	dbg_init();
	dbg_printf("programstart:\n");
  #endif
	uint32_t oldpad = 0, curpad = 0;
	int32_t ret;
  CellPadData pdata;

  #ifdef HAVE_STARFIELD
  init_once(/* stars */);
  #endif

  // wait for XMB feedback
  sys_timer_sleep(13);
	vshtask_notify("PS3 VSH Menu Loaded\nPress [Select] to open it");
  auto_game_load = 1;
  sprintf(current_auto_setting, "ENABLED");
  plugin_running = IN_VSH;
  while(plugin_running)
	{
    if (plugin_running == IN_VSH)
    {
      if(!menu_running)                                                   // VSH menu is not running, normal XMB execution
      {
        VSHPadGetData(&pdata);                                            // use VSHPadGetData() to check pad
        if(pdata.len > 0)                                                 // if pad data
        {
          curpad = (pdata.button[2] | (pdata.button[3] << 8));            // merge pad data
          if(curpad != oldpad)
          {
            if(curpad & PAD_SELECT)
            {
              if (curpad & PAD_START)
              {
                if (curpad & PAD_R3)                                      // if select+start+R3 was pressed, load game menu plugin manually
                {
                  if(vshmain_EB757101() != 0)                             // if game has booted
                  {
                    check_for_booted_game();
                  }
                }
                else if(curpad & PAD_R1)
                {
                  if(vshmain_EB757101() == 0)
                  {
                    screenshot(0);
                  }
                }
              }
              else if(!(curpad & PAD_START))                              // if select was pressed, start VSH menu
              {
                if(vshmain_EB757101() == 0)                               // if we are in XMB
                {
                  ret = start_VSH_Menu();
                  if (ret != 0)
                  {
                    stop_VSH_Menu();                                      // failed to init graphics, try pressing select again
                  }
                }
              }
            }
          }
          oldpad = curpad;
        }
        else
        {
          oldpad = 0;
        }
        if (auto_game_load)
        {
          check_for_booted_game();                                        // check if game is booted
        }
        sys_timer_usleep(70000);                                          // vsh sync
      }
      else                                                                // menu is running, draw frame, flip frame, check for pad events, sleep, ...
      {
        draw_frame();
        flip_frame();
        MyPadGetData(0, &pdata);                                          // use MyPadGetData() during VSH menu
        if(pdata.len > 0)                                                 // if pad data
        {
          curpad = (pdata.button[2] | (pdata.button[3] << 8));            // merge pad data
          if(curpad != oldpad)
          {
            if(curpad & PAD_SELECT)
            {
              if (curpad & PAD_START)
              {
                screenshot(1);
              }
              else if(!(curpad & PAD_START))                              // if select was pressed stop VSH menu
              {
                stop_VSH_Menu();
                sys_timer_usleep(100000);
              }
            }
            else if(curpad & PAD_UP)
            {
              if(line > 0)
              {
                line--;
              }
              else
                line = 0;
            }
            else if(curpad & PAD_DOWN)
            {
              if(line < (view_menu[view]-1))
              {
                line++;
              }
              else
                line = (view_menu[view]-1);
            }
            else if(curpad & PAD_CROSS)
            {
              do_menu_action();
            }
            oldpad = curpad;
          }
        }
        else
        {
          oldpad = 0;
        }
        sys_timer_usleep(200);                                          // short menu frame delay
      }
    }
    else                                                                // IN_GAME
    {
      check_for_exited_game();
      sys_timer_usleep(100000);
    }
	}

  // unloading plugin...
	vshtask_notify("PS3 VSH Menu Unloaded");
  #ifdef DEBUG
	dbg_fini();
  #endif
  return(0);
}

////////////////////////////////////////////////////////////////////////
// VSH PLUGIN SPECIFIC METHODS
/***********************************************************************
* check if a game has booted and load the game menu plugin to it
***********************************************************************/
void check_for_booted_game(void)
{
  int32_t r;

  if (FindLoadedPlugin("game_plugin") != 0) {

    if (auto_game_load == 1) {
      // wait for game feedback
      sys_timer_sleep(13);
    }
    r = load_plugin_to_game("/dev_hdd0/plugins/ps3_menu/ps3_game_menu.sprx", 0, 0);
    if (r == 0) {
      plugin_running = IN_GAME;
	    vshtask_notify("PS3 Game Menu Loaded\nPress [Select] to open it");
    }
    else
    {
	    vshtask_notify("PS3 Game Menu Failed to Load\nPress [Select+Start+R3] to manually load it");
    }
  }
}

/***********************************************************************
* check if a game has exited and return control back to vsh menu
***********************************************************************/
void check_for_exited_game(void)
{
  if (FindLoadedPlugin("game_plugin") == 0) {
    plugin_running = IN_VSH;
  }
}

/***********************************************************************
* load a game plugin using PS3MAPI interface
***********************************************************************/
int32_t load_plugin_to_game(uint8_t *prx_path, void *arg, uint32_t arg_size)
{
  int32_t r;
  char pid_str[128];
  uint32_t pid_list[16];
  sys_prx_id_t game_pid = 0;

  {system_call_3(8, SYSCALL8_OPCODE_PS3MAPI, PS3MAPI_OPCODE_GET_ALL_PROC_PID, (uint64_t)((uint32_t)pid_list));}
  for(int i = 0; i < 16; i++) {
    if (1 < pid_list[i]) {
      memset(pid_str, 0, sizeof(pid_str));
      {system_call_4(8, SYSCALL8_OPCODE_PS3MAPI, PS3MAPI_OPCODE_GET_PROC_NAME_BY_PID, (uint64_t)((uint32_t)pid_list[i]), (uint64_t)((uint32_t)pid_str));}
      if (strstr(pid_str, "EBOOT")) {
        game_pid = pid_list[i];
        break;
      } else if ((strstr(pid_str, "vsh") == NULL) && (strstr(pid_str, "agent") == NULL) && (strstr(pid_str, "sys") == NULL)) {

        // in case it's a self
        game_pid = pid_list[i];
        break;
      }
    }
  }
  {
    system_call_6(8, SYSCALL8_OPCODE_PS3MAPI, PS3MAPI_OPCODE_LOAD_PROC_MODULE, (uint64_t)game_pid, (uint64_t)((uint32_t)prx_path), (uint64_t)((uint32_t)arg), (uint64_t)((uint32_t)arg_size));
    r = (int32_t)p1;
  }
  return(r);
}

